
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.23.11.post1',
    'branch_nick': 'hotfix-aws',
    'revision_id': '1811d23d84f8d26f982e216f144c77460db7e885',
    'revno': 10085
}
